
/* $RCSfile: TclImport.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:17 $
 *
 * Copyright (C) 1995-2003 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "tcl.h"
#include "Qddb.h"
#include "tclQddb.h"

/* qddb_import cvs <string>
 */

static int ImportCSV _ANSI_ARGS_((Tcl_Interp *, int, CONST char **));
static int CSVRowCount _ANSI_ARGS_((Qddb_Entry));
static int CSVColCount _ANSI_ARGS_((char *));
static int CSVImportRow _ANSI_ARGS_((Qddb_Table *, int, char *, int *));
static int CSVSetCellValue _ANSI_ARGS_((Qddb_TableNode *, char *, int));
static void SetColumnTypes _ANSI_ARGS_((Qddb_Table *));

int TclQddb_Import(clientData, interp, argc, argv)
    ClientData			clientData;
    Tcl_Interp			*interp;
    int				argc;
    CONST char			**argv;
{
    if (argc != 3) {
	Tcl_AppendResult(interp, argv[0], ": wrong # args", NULL);
	return TCL_ERROR;
    }
    if (strcmp(argv[1], "csv") == 0) {
	ImportCSV(interp, argc, argv);
    } else {
	Tcl_AppendResult(interp, argv[0], ": bad option \"", argv[1], "\"", NULL);
	return TCL_ERROR;
    }
    return TCL_OK;
}

static int CSVRowCount(lines)
    Qddb_Entry                  lines;
{
    int                         retval;

    retval = 0;
    while (*lines++ != NULL)
	retval++;
    return retval;
}

static int CSVColCount(line)
    char                        *line;
{
    char                        prev, *start_char;
    int                         inquote, column;

    inquote = 0;
    column = 1;
    start_char = line;
    while (*line != '\0') {
	prev = *line;
	switch (*line) {
	case '\n':
	case '\r':
	    break;
	case ',':
	    if (!inquote) {
		column++;
		start_char = line+1;
	    }
	    break;
	case '"':
	    if (prev != '\\') {
		if (!inquote) {
		    start_char = line+1;
		    inquote = 1;
		} else {
		    inquote = 0;
		}
	    }
	    break;
	default:
	    break;
	}
	line++;
    }
    return column;
}

static int CSVSetCellValue(cell, value, columntype)
    Qddb_TableNode		*cell;
    char			*value;
    int                         columntype;
{
    size_t                      len;

    len = strlen(value);
    cell->data.type = QDDB_TABLE_TYPE_DEFAULT;
    if (Qddb_IsNumericString(value) && len > 0) {
	cell->data.data.real = strtod(value, NULL);
	cell->data.type = QDDB_TABLE_TYPE_REAL;
	if (floor(cell->data.data.real) == cell->data.data.real) {
	    cell->data.data.integer = (int)cell->data.data.real;
	    cell->data.type = QDDB_TABLE_TYPE_INTEGER;
	}
    } else if ((columntype == QDDB_TABLE_TYPE_DEFAULT || 
	       columntype == QDDB_TABLE_TYPE_DATE) &&
	       len > 0 &&
	       Qddb_CheckDate(value) == 0) {
	cell->data.type = QDDB_TABLE_TYPE_DATE;
	cell->data.data.string = (char *)Malloc(strlen(value)+1);
	strcpy(cell->data.data.string, value);	
    }
    if (cell->data.type == QDDB_TABLE_TYPE_DEFAULT) {
	if (len > 0) {
	    cell->data.data.string = (char *)Malloc(len+1);
	    strcpy(cell->data.data.string, value);
	}
	cell->data.type = QDDB_TABLE_TYPE_STRING;
    }
    return cell->data.type;
}


static int CSVImportRow(table, row, line, columntypes)
    Qddb_Table                  *table;
    int                         row;
    char                        *line;
    int                         *columntypes;
{
    char                        prev, *start_char, *orig_line;
    int                         inquote, column;

    inquote = 0;
    column = 1;
    orig_line = start_char = line;
    prev = '\0';
    while (*line != '\0') {
	switch (*line) {
	case '\n':
	case '\r':
	    break;
	case ',':
	    if (!inquote) {
		if (column > table->columns) {
		    return column;
		}
		if (prev == '"') {
		    *(line-1) = '\0';
		} else {
		    *line = '\0';
		}
		columntypes[column] = CSVSetCellValue(table->table[row] + column, start_char, 
						      columntypes[column]);
		if (prev == '"') {
		    *(line-1) = '"';
		} else {
		    *line = ',';
		}
		column++;
		start_char = line+1;
	    }
	    break;
	case '"':
	    if (prev != '\\') {
		if (!inquote) {
		    start_char = line+1;
		    inquote = 1;
		} else {
		    inquote = 0;
		}
	    }
	    break;
	default:
	    break;
	}
	prev = *line;
	line++;
    }
    if (orig_line != line) {
	if (column > table->columns) {
	    if (column == table->columns + 1) {
		return 0;
	    } else {
		return column - 1;
	    }
	}
	if (prev == '"') {
	    *(line-1) = '\0';
	} else {
	    *line = '\0';
	}
	columntypes[column] = CSVSetCellValue(table->table[row] + column, start_char,
					      columntypes[column]);
	if (prev == '"') {
	    *(line-1) = '"';
	} else {
	    *line = ',';
	}
    }
    return 0;
}

static int ImportCSV(interp, argc, argv)
    Tcl_Interp                  *interp;
    int                         argc;
    CONST char                  **argv;
{
    Qddb_Table                  *table_ptr;
    Qddb_TableRowCol            rowcol;
    Qddb_Entry                  lines, orig_lines;
    char                        *table_desc, *buffer;
    char                        perfval[BUFSIZ];
    CONST char                  *val;
    int                         i, *columntypes, perfmeter;

    lines = NULL;
    buffer = Malloc(strlen(argv[2])+1);
    strcpy(buffer, argv[2]);
    Qddb_SplitBufferIntoLines(&lines, buffer, strlen(argv[2]));
    orig_lines = lines;
    rowcol.row = CSVRowCount(lines);
    rowcol.column = CSVColCount(*lines);
#if defined(DEBUG)
    fprintf(stderr, "rows=%d, cols=%d\n", rowcol.row, rowcol.column);
    fflush(stderr);
#endif
    table_ptr = Qddb_TableOp(NULL, QDDB_TABLE_OP_DEFINE, (void *)&rowcol);
    if (table_ptr == NULL) {
	Tcl_AppendResult(interp, "qddb_table define failed", NULL);
	return TCL_ERROR;
    }
    columntypes = (int *)Calloc(sizeof(int)*(rowcol.column+1));
    if ((val = Tcl_GetVar(interp, "qddb_perfmeter", TCL_GLOBAL_ONLY)) == NULL) {
	perfmeter = 1;
    } else {
	perfmeter = strtol(val, NULL, 10);
    }
    for (i = 1; *lines != NULL && i <= rowcol.row; i++, lines++) {
#if defined(DEBUG)
	int columns;
	if ((columns=CSVImportRow(table_ptr, i, (char *)*lines, columntypes)) != 0) {
	    fprintf(stderr, "wrong # columns (%d should be %d)\n", columns, rowcol.column);
	}
#else
	(void)CSVImportRow(table_ptr, i, (char *)*lines, columntypes);
#endif
	sprintf(perfval, "%d", i + perfmeter);
	Tcl_SetVar(interp, "qddb_perfmeter", perfval, TCL_GLOBAL_ONLY);
	while (Tcl_DoOneEvent(TCL_DONT_WAIT));
#if defined(DEBUG)
	fprintf(stderr, "Imported: %s\n", *lines);
	fflush(stderr);
#endif
    }
    Qddb_Free(QDDB_TYPE_ENTRY, orig_lines);
    Free(columntypes);
#if defined(notyet)
    SetColumnTypes(table_ptr);
#endif
    if ((table_desc = TclQddb_NewTable(interp, table_ptr)) == NULL) {
	Tcl_AppendResult(interp, argv[0], argv[1], ": cannot create table", NULL);
	return TCL_ERROR;
    }
    Tcl_SetResult(interp, table_desc, TCL_DYNAMIC);
    return TCL_OK;
}

static void SetColumnTypes(table)
    Qddb_Table                  *table;
{
    Qddb_TableNode              *row;
    int                         i, j;

    for (i = 1; i <= table->rows; i++) {
	row = table->table[i];
	for (j = 1; j <= table->columns; j++) {
	    if (row[j].data.type == QDDB_TABLE_TYPE_STRING) {
		break;
	    } else if (row[j].data.type == QDDB_TABLE_TYPE_REAL) {
		break;
	    }
	}
    }
}





