
/* $RCSfile: TclLink.c,v $ $Revision: 2.1 $ $Date: 2003/03/23 17:16:17 $
 *
 * Copyright (C) 1995-2003 Herrin Software Development, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* 
 * tclLink.c --
 *
 *	This file implements linked variables (a C variable that is
 *	tied to a Tcl variable).  The idea of linked variables was
 *	first suggested by Andreas Stolcke and this implementation is
 *	based heavily on a prototype implementation provided by
 *	him.
 *
 * Copyright (c) 1993 The Regents of the University of California.
 * Copyright (c) 1994 Sun Microsystems, Inc.
 *
 * This software is copyrighted by the Regents of the University of
 * California, Sun Microsystems, Inc., and other parties.  The following
 * terms apply to all files associated with the software unless explicitly
 * disclaimed in individual files.
 * 
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 * 
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
 * DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
 * IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
 * NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 * 
 * RESTRICTED RIGHTS: Use, duplication or disclosure by the government
 * is subject to the restrictions as set forth in subparagraph (c) (1) (ii)
 * of the Rights in Technical Data and Computer Software Clause as DFARS
 * 252.227-7013 and FAR 52.227-19.
 */

#include "Qddb.h"
#include "tclQddb.h"

/*
 * For each linked variable there is a data structure of the following
 * type, which describes the link and is the clientData for the trace
 * set on the Tcl variable.
 */

typedef struct qddb_Link {
    Tcl_Interp    *interp;	/* Interpreter containing Tcl variable. */
    char          *varName;	/* Name of variable (must be global).  This
				 * is needed during trace callbacks, since
				 * the actual variable may be aliased at
				 * that time via upvar. */
    char          *addr;	/* Location of C variable. */
    char          *format;	/* Format of the string from the Schema */
    Qddb_DataTree *node;	/* The node linked */
    int           type;		/* Type of link (TCL_LINK_INT, etc.). */
    int           flags;
    union {
        int       i;
	double    d;
    } lastValue;		/* Last known value of C variable;  used to
				 * avoid string conversions. */
} Qddb_Link;

#undef LINK_READ_ONLY
#undef LINK_BEING_UPDATED
#define LINK_READ_ONLY		0x01
#define LINK_BEING_UPDATED	0x02

/*
 * Forward references to procedures defined later in this file:
 */
static Tcl_VarTraceProc     LinkTraceProc;
static char *		    StringValue _ANSI_ARGS_((Qddb_Link *linkPtr,
			    char *buffer));

/*
 *----------------------------------------------------------------------
 *
 * TclQddb_LinkVar --
 *
 *	Link a C variable to a Tcl variable so that changes to either
 *	one causes the other to change.
 *
 * Results:
 *	The return value is TCL_OK if everything went well or TCL_ERROR
 *	if an error occurred (interp->result is also set after errors).
 *
 * Side effects:
 *	The value at *addr is linked to the Tcl variable "varName",
 *	using "type" to convert between string values for Tcl and
 *	binary values for *addr.
 *
 *----------------------------------------------------------------------
 */

int TclQddb_LinkVar(interp, varName, addr, type, format, node)
    Tcl_Interp *interp;		/* Interpreter in which varName exists. */
    char *varName;		/* Name of a global variable in interp. */
    char *addr;			/* Address of a C variable to be linked
				 * to varName. */
    int type;			/* Type of C variable: TCL_LINK_INT, etc. 
				 * Also may have TCL_LINK_READ_ONLY
				 * OR'ed in. */
    char *format;		/* variable format */
    Qddb_DataTree *node;	/* node */
{
    Qddb_Link *linkPtr;
    char buffer[BUFSIZ];
    int code;

    linkPtr = (Qddb_Link *) Malloc(sizeof(Qddb_Link));
    linkPtr->interp = interp;
    linkPtr->varName = Malloc(strlen(varName)+1);
    strcpy(linkPtr->varName, varName);
    linkPtr->format = format;
    linkPtr->addr = addr;
    linkPtr->type = type & ~TCL_LINK_READ_ONLY;
    if (type & TCL_LINK_READ_ONLY) {
	linkPtr->flags = LINK_READ_ONLY;
    } else {
	linkPtr->flags = 0;
    }
    linkPtr->node = node;
    if (Tcl_SetVar(interp, varName, StringValue(linkPtr, buffer),
	    TCL_GLOBAL_ONLY|TCL_LEAVE_ERR_MSG) == NULL) {
	Free(linkPtr->varName);
	Free((char *) linkPtr);
	return TCL_ERROR;
    }
    code = Tcl_TraceVar(interp, varName, TCL_GLOBAL_ONLY|TCL_TRACE_READS
	    |TCL_TRACE_WRITES|TCL_TRACE_UNSETS, LinkTraceProc,
	    (ClientData) linkPtr);
    if (code != TCL_OK) {
	Free(linkPtr->varName);
	Free((char *) linkPtr);
    }
    return code;
}

/*
 *----------------------------------------------------------------------
 *
 * TclQddb_UnlinkVar --
 *
 *	Destroy the link between a Tcl variable and a C variable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If "varName" was previously linked to a C variable, the link
 *	is broken to make the variable independent.  If there was no
 *	previous link for "varName" then nothing happens.
 *
 *----------------------------------------------------------------------
 */

void
TclQddb_UnlinkVar(interp, varName)
    Tcl_Interp *interp;		/* Interpreter containing variable to unlink. */
    char *varName;		/* Global variable in interp to unlink. */
{
    Qddb_Link *linkPtr;

    linkPtr = (Qddb_Link *) Tcl_VarTraceInfo(interp, varName, TCL_GLOBAL_ONLY,
	    LinkTraceProc, (ClientData) NULL);
    if (linkPtr == NULL) {
	return;
    }
    if (linkPtr->type == TCL_LINK_STRING) {
	CONST char 	*value;
        char            *retval;
	char		**place;

	Tcl_UntraceVar(interp, varName,
		       TCL_GLOBAL_ONLY|TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		       LinkTraceProc, (ClientData) linkPtr);
	place = (char **)(linkPtr->addr);
 	value = Tcl_GetVar(interp, varName, TCL_GLOBAL_ONLY);
	retval = Malloc(strlen(value)+1);
	strcpy(retval, value);
	if (*place != NULL)
	    Free(*place);
	*place = retval;
    } else {
	Tcl_UntraceVar(interp, varName,
		       TCL_GLOBAL_ONLY|TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		       LinkTraceProc, (ClientData) linkPtr);
    }
    Free(linkPtr->varName);
    Free((char *) linkPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TclQddb_UpdateLinkedVar --
 *
 *	This procedure is invoked after a linked variable has been
 *	changed by C code.  It updates the Tcl variable so that
 *	traces on the variable will trigger.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The Tcl variable "varName" is updated from its C value,
 *	causing traces on the variable to trigger.
 *
 *----------------------------------------------------------------------
 */

void
TclQddb_UpdateLinkedVar(interp, varName)
    Tcl_Interp       *interp;		/* Interpreter containing variable. */
    char             *varName;		/* Name of global variable that is linked. */
{
    Qddb_Link        *linkPtr;
    char             buffer[TCL_DOUBLE_SPACE];
    int              savedFlag;

    linkPtr = (Qddb_Link *)Tcl_VarTraceInfo(interp, varName, TCL_GLOBAL_ONLY,
					    LinkTraceProc, (ClientData) NULL);
    if (linkPtr == NULL) {
	return;
    }
    savedFlag = linkPtr->flags & LINK_BEING_UPDATED;
    linkPtr->flags |= LINK_BEING_UPDATED;
    Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
    linkPtr->flags = (linkPtr->flags & ~LINK_BEING_UPDATED) | savedFlag;
}

/*
 *----------------------------------------------------------------------
 *
 * LinkTraceProc --
 *
 *	This procedure is invoked when a linked Tcl variable is read,
 *	written, or unset from Tcl.  It's responsible for keeping the
 *	C variable in sync with the Tcl variable.
 *
 * Results:
 *	If all goes well, NULL is returned; otherwise an error message
 *	is returned.
 *
 * Side effects:
 *	The C variable may be updated to make it consistent with the
 *	Tcl variable, or the Tcl variable may be overwritten to reject
 *	a modification.
 *
 *----------------------------------------------------------------------
 */

static char *
LinkTraceProc(clientData, interp, name1, name2, flags)
    ClientData clientData;	/* Contains information about the link. */
    Tcl_Interp *interp;		/* Interpreter containing Tcl variable. */
    CONST84 char *name1;	/* First part of variable name. */
    CONST84 char *name2;        /* Second part of variable name. */
    int flags;			/* Miscellaneous additional information. */
{
    Qddb_Link   *linkPtr = (Qddb_Link *) clientData;
    int         changed;
    char        buffer[BUFSIZ];
    CONST char  *value;
    char        **pp;
    Tcl_DString savedResult;

    /*
     * If the variable is being unset, then just re-create it (with a
     * trace) unless the whole interpreter is going away.
     */

    if (flags & TCL_TRACE_UNSETS) {
	if (flags & TCL_INTERP_DESTROYED) {
	    Free(linkPtr->varName);
	    Free((char *) linkPtr);
	} else if (flags & TCL_TRACE_DESTROYED) {
	    Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
		    TCL_GLOBAL_ONLY);
	    Tcl_TraceVar(interp, linkPtr->varName, TCL_GLOBAL_ONLY
		    |TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS,
		    LinkTraceProc, (ClientData) linkPtr);
	}
	return NULL;
    }
    if (linkPtr->flags & LINK_BEING_UPDATED) {
	return NULL;
    }

    /*
     * For read accesses, update the Tcl variable if the C variable
     * has changed since the last time we updated the Tcl variable.
     */

    if (flags & TCL_TRACE_READS) {
	switch (linkPtr->type) {
	    case TCL_LINK_INT:
	    case TCL_LINK_BOOLEAN:
		changed = *(int *)(linkPtr->addr) != linkPtr->lastValue.i ||
		    linkPtr->node->modified == False;
		break;
	    case TCL_LINK_DOUBLE:
		changed = *(double *)(linkPtr->addr) != linkPtr->lastValue.d ||
		    linkPtr->node->modified == False;
		break;
	    case TCL_LINK_STRING:
		changed = 1;
		break;
	    default:
		return "internal error: bad linked variable type";
	}
	if (changed) {
	    Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
		    TCL_GLOBAL_ONLY);
	}
	return NULL;
    }

    /*
     * For writes, first make sure that the variable is writable.  Then
     * convert the Tcl value to C if possible.  If the variable isn't
     * writable or can't be converted, then restore the varaible's old
     * value and return an error.  Another tricky thing: we have to save
     * and restore the interpreter's result, since the variable access
     * could occur when the result has been partially set.
     */
    if ((linkPtr->flags&LINK_READ_ONLY) != 0) {
	Tcl_SetVar(interp, linkPtr->varName, StringValue(linkPtr, buffer),
		TCL_GLOBAL_ONLY);
	return "linked variable is read-only";
    }
    value = Tcl_GetVar(interp, linkPtr->varName, TCL_GLOBAL_ONLY);
    if (value == NULL) {
	/*
	 * This shouldn't ever happen.
	 */
	return "internal error: linked variable couldn't be read";
    }
    Tcl_DStringInit(&savedResult);
    Tcl_DStringAppend(&savedResult, interp->result, -1);
    Tcl_ResetResult(interp);
    switch (linkPtr->type) {
    case TCL_LINK_INT:
	if (value[0] == '\0') {
	    linkPtr->node->modified = False;
	    linkPtr->lastValue.i = 0;
	    break;
	} else if (Tcl_GetInt(interp, value, &linkPtr->lastValue.i) != TCL_OK) {
	    Tcl_DStringResult(interp, &savedResult);
	    Tcl_SetVar(interp, linkPtr->varName,
		       StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
	    return "variable must have integer value";
	} else {
	    linkPtr->node->modified = True;
	}
	*(int *)(linkPtr->addr) = linkPtr->lastValue.i;
	break;
    case TCL_LINK_DOUBLE:
	if (value[0] == '\0') {
	    linkPtr->node->modified = False;
	    linkPtr->lastValue.d = 0.00;
	} else if (Tcl_GetDouble(interp, value, &linkPtr->lastValue.d)
	    != TCL_OK) {
	    Tcl_DStringResult(interp, &savedResult);
	    Tcl_SetVar(interp, linkPtr->varName,
		       StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
	    return "variable must have real value";
	} else {
	    linkPtr->node->modified = True;
	}
	*(double *)(linkPtr->addr) = linkPtr->lastValue.d;
	break;
    case TCL_LINK_BOOLEAN:
	if (Tcl_GetBoolean(interp, value, &linkPtr->lastValue.i)
	    != TCL_OK) {
	    Tcl_DStringResult(interp, &savedResult);
	    Tcl_SetVar(interp, linkPtr->varName,
		       StringValue(linkPtr, buffer), TCL_GLOBAL_ONLY);
	    return "variable must have boolean value";
	}
	*(int *)(linkPtr->addr) = linkPtr->lastValue.i;
	break;
    case TCL_LINK_STRING:
	pp = (char **)(linkPtr->addr);
	if (*pp != NULL) {
	    Free(*pp);
	}
	*pp = Malloc(strlen(value)+1);
	strcpy(*pp, value);
	linkPtr->node->modified = True;
	break;
    default:
	return "internal error: bad linked variable type";
    }
    Tcl_DStringResult(interp, &savedResult);
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * StringValue --
 *
 *	Converts the value of a C variable to a string for use in a
 *	Tcl variable to which it is linked.
 *
 * Results:
 *	The return value is a pointer
 to a string that represents
 *	the value of the C variable given by linkPtr.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static char *
StringValue(linkPtr, buffer)
    Qddb_Link *linkPtr;		/* Structure describing linked variable. */
    char *buffer;		/* Small buffer to use for converting
				 * values.  Must have TCL_DOUBLE_SPACE
				 * bytes or more. */
{
    char *p;

    switch (linkPtr->type) {
	case TCL_LINK_INT:
	    linkPtr->lastValue.i = *(int *)(linkPtr->addr);
	    if (linkPtr->format != NULL) {
		sprintf(buffer, linkPtr->format, linkPtr->lastValue.i);		
	    } else {
		sprintf(buffer, "%d", linkPtr->lastValue.i);
	    }
	    if (linkPtr->node->modified == True) {
		return buffer;
	    } else {
		buffer[0] = '\0';
		return buffer;
	    }
	case TCL_LINK_DOUBLE:
	    linkPtr->lastValue.d = *(double *)(linkPtr->addr);
	    if (linkPtr->format != NULL) {
		sprintf(buffer, linkPtr->format, linkPtr->lastValue.d);
	    } else {
		sprintf(buffer, "%f", linkPtr->lastValue.d);
	    }
	    if (linkPtr->node->modified == True) {
		return buffer;
	    } else {
		buffer[0] = '\0';
		return buffer;
	    }
	case TCL_LINK_BOOLEAN:
	    linkPtr->lastValue.i = *(int *)(linkPtr->addr);
	    if (linkPtr->lastValue.i != 0) {
		return "1";
	    }
	    return "0";
	case TCL_LINK_STRING:
	    p = *(char **)(linkPtr->addr);
	    if (p == NULL) {
		return "NULL";
	    }
	    return p;
    }

    /*
     * This code only gets executed if the link type is unknown
     * (shouldn't ever happen).
     */

    return "??";
}
